@@ -1,36 +1,66 @@
-# Import stuff
+"""Hook Points.
+
+Helpers to access activations in models.
+"""
+
 import logging
 from contextlib import contextmanager
 from dataclasses import dataclass
-from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union
+from functools import partial
+from typing import (
+    Any,
+    Callable,
+    Dict,
+    Iterable,
+    List,
+    Literal,
+    Optional,
+    Protocol,
+    Sequence,
+    Tuple,
+    Union,
+    runtime_checkable,
+)
 
+import torch
 import torch.nn as nn
 import torch.utils.hooks as hooks
 
+from transformer_lens.utils import Slice, SliceInput
+
 
 @dataclass
 class LensHandle:
-    """
-    A dataclass that holds information about a PyTorch hook.
-
-    Attributes:
-        hook (hooks.RemovableHandle): Reference to the hook's RemovableHandle.
-        is_permanent (bool, optional): Indicates if the hook is permanent. Defaults to False.
-        context_level (Optional[int], optional): Context level associated with the hooks context
-            manager for the given hook. Defaults to None.
-    """
+    """Dataclass that holds information about a PyTorch hook."""
 
     hook: hooks.RemovableHandle
+    """Reference to the Hook's Removable Handle."""
+
     is_permanent: bool = False
+    """Indicates if the Hook is Permanent."""
+
     context_level: Optional[int] = None
+    """Context level associated with the hooks context manager for the given hook."""
 
 
-# %%
 # Define type aliases
 NamesFilter = Optional[Union[Callable[[str], bool], Sequence[str]]]
 
 
-# %%
+@runtime_checkable
+class _HookFunctionProtocol(Protocol):
+    """Protocol for hook functions."""
+
+    def __call__(self, tensor: torch.Tensor, *, hook: "HookPoint") -> Union[Any, None]:
+        ...
+
+
+HookFunction = _HookFunctionProtocol  # Callable[..., _HookFunctionProtocol]
+
+DeviceType = Optional[torch.device]
+_grad_t = Union[Tuple[torch.Tensor, ...], torch.Tensor]
+
+
 class HookPoint(nn.Module):
     """
     A helper class to access intermediate activations in a PyTorch model (inspired by Garcon).
@@ -47,13 +77,18 @@ class HookPoint(nn.Module):
 
         # A variable giving the hook's name (from the perspective of the root
         # module) - this is set by the root module at setup.
-        self.name = None
+        self.name: Union[str, None] = None
 
-    def add_perma_hook(self, hook, dir="fwd") -> None:
+    def add_perma_hook(self, hook: HookFunction, dir: Literal["fwd", "bwd"] = "fwd") -> None:
         self.add_hook(hook, dir=dir, is_permanent=True)
 
     def add_hook(
-        self, hook, dir="fwd", is_permanent=False, level=None, prepend=False
+        self,
+        hook: HookFunction,
+        dir: Literal["fwd", "bwd"] = "fwd",
+        is_permanent: bool = False,
+        level: Optional[int] = None,
+        prepend: bool = False,
     ) -> None:
         """
         Hook format is fn(activation, hook_name)
@@ -61,57 +96,55 @@ class HookPoint(nn.Module):
         which are the same for a HookPoint)
         If prepend is True, add this hook before all other hooks
         """
+
+        def full_hook(
+            module: torch.nn.Module,
+            module_input: Any,
+            module_output: Any,
+        ):
+            if (
+                dir == "bwd"
+            ):  # For a backwards hook, module_output is a tuple of (grad,) - I don't know why.
+                module_output = module_output[0]
+            return hook(module_output, hook=self)
+
+        full_hook.__name__ = (
+            hook.__repr__()
+        )  # annotate the `full_hook` with the string representation of the `hook` function
+
         if dir == "fwd":
-
-            def full_hook(module, module_input, module_output):
-                return hook(module_output, hook=self)
-
-            full_hook.__name__ = (
-                hook.__repr__()
-            )  # annotate the `full_hook` with the string representation of the `hook` function
-
-            handle = self.register_forward_hook(full_hook)
-            handle = LensHandle(handle, is_permanent, level)
-
-            if prepend:
-                # we could just pass this as an argument in PyTorch 2.0, but for now we manually do this...
-                self._forward_hooks.move_to_end(handle.hook.id, last=False)
-                self.fwd_hooks.insert(0, handle)
-
-            else:
-                self.fwd_hooks.append(handle)
-
+            pt_handle = self.register_forward_hook(full_hook)
+            _internal_hooks = self._forward_hooks
+            visible_hooks = self.fwd_hooks
         elif dir == "bwd":
-            # For a backwards hook, module_output is a tuple of (grad,) - I don't know why.
-
-            def full_hook(module, module_input, module_output):
-                return hook(module_output[0], hook=self)
-
-            full_hook.__name__ = (
-                hook.__repr__()
-            )  # annotate the `full_hook` with the string representation of the `hook` function
-
-            handle = self.register_full_backward_hook(full_hook)
-            handle = LensHandle(handle, is_permanent, level)
-
-            if prepend:
-                # we could just pass this as an argument in PyTorch 2.0, but for now we manually do this...
-                self._backward_hooks.move_to_end(handle.hook.id, last=False)
-                self.bwd_hooks.insert(0, handle)
-            else:
-                self.bwd_hooks.append(handle)
+            pt_handle = self.register_backward_hook(full_hook)
+            _internal_hooks = self._backward_hooks
+            visible_hooks = self.bwd_hooks
         else:
             raise ValueError(f"Invalid direction {dir}")
 
-    def remove_hooks(self, dir="fwd", including_permanent=False, level=None) -> None:
+        handle = LensHandle(pt_handle, is_permanent, level)
+
+        if prepend:
+            # we could just pass this as an argument in PyTorch 2.0, but for now we manually do this...
+            _internal_hooks.move_to_end(handle.hook.id, last=False)  # type: ignore # TODO: this type error could signify a bug
+            visible_hooks.insert(0, handle)
+
+        else:
+            visible_hooks.append(handle)
+
+    def remove_hooks(
+        self,
+        dir: Literal["fwd", "bwd", "both"] = "fwd",
+        including_permanent: bool = False,
+        level: Optional[int] = None,
+    ) -> None:
         def _remove_hooks(handles: List[LensHandle]) -> List[LensHandle]:
             output_handles = []
             for handle in handles:
                 if including_permanent:
                     handle.hook.remove()
-                elif (not handle.is_permanent) and (
-                    level is None or handle.context_level == level
-                ):
+                elif (not handle.is_permanent) and (level is None or handle.context_level == level):
                     handle.hook.remove()
                 else:
                     output_handles.append(handle)
@@ -128,29 +161,45 @@ class HookPoint(nn.Module):
         del self.ctx
         self.ctx = {}
 
-    def forward(self, x):
+    def forward(self, x: torch.Tensor) -> torch.Tensor:
         return x
 
     def layer(self):
         # Returns the layer index if the name has the form 'blocks.{layer}.{...}'
         # Helper function that's mainly useful on HookedTransformer
         # If it doesn't have this form, raises an error -
+        if self.name is None:
+            raise ValueError("Name cannot be None")
         split_name = self.name.split(".")
         return int(split_name[1])
 
 
 # %%
 class HookedRootModule(nn.Module):
-    """
-    A class building on nn.Module to interface nicely with HookPoints
-    Adds various nice utilities, most notably run_with_hooks to run the model with temporary hooks, and run_with_cache to run the model on some input and return a cache of all activations
+    """A class building on nn.Module to interface nicely with HookPoints.
 
-    WARNING: The main footgun with PyTorch hooking is that hooks are GLOBAL state. If you add a hook to the module, and then run it a bunch of times, the hooks persist. If you debug a broken hook and add the fixed version, the broken one is still there. To solve this, run_with_hooks will remove hooks at the end by default, and I recommend using the API of this and run_with_cache. If you want to add hooks into global state, I recommend being intentional about this, and I recommend using reset_hooks liberally in your code to remove any accidentally remaining global state.
+    Adds various nice utilities, most notably run_with_hooks to run the model with temporary hooks,
+    and run_with_cache to run the model on some input and return a cache of all activations.
 
-    The main time this goes wrong is when you want to use backward hooks (to cache or intervene on gradients). In this case, you need to keep the hooks around as global state until you've run loss.backward() (and so need to disable the reset_hooks_end flag on run_with_hooks)
+    Notes:
+
+    The main footgun with PyTorch hooking is that hooks are GLOBAL state. If you add a hook to the
+    module, and then run it a bunch of times, the hooks persist. If you debug a broken hook and add
+    the fixed version, the broken one is still there. To solve this, run_with_hooks will remove
+    hooks at the end by default, and I recommend using the API of this and run_with_cache. If you
+    want to add hooks into global state, I recommend being intentional about this, and I recommend
+    using reset_hooks liberally in your code to remove any accidentally remaining global state.
+
+    The main time this goes wrong is when you want to use backward hooks (to cache or intervene on
+    gradients). In this case, you need to keep the hooks around as global state until you've run
+    loss.backward() (and so need to disable the reset_hooks_end flag on run_with_hooks)
     """
 
-    def __init__(self, *args):
+    name: Optional[str]
+    mod_dict: Dict[str, nn.Module]
+    hook_dict: Dict[str, HookPoint]
+
+    def __init__(self, *args: Any):
         super().__init__()
         self.is_caching = False
         self.context_level = 0
@@ -165,25 +214,27 @@ class HookedRootModule(nn.Module):
         "HookPoint".
         """
         self.mod_dict = {}
-        self.hook_dict: Dict[str, HookPoint] = {}
+        self.hook_dict = {}
         for name, module in self.named_modules():
             if name == "":
                 continue
             module.name = name
             self.mod_dict[name] = module
-            if "HookPoint" in str(type(module)):
+            # TODO: is the bottom line the same as "if "HookPoint" in str(type(module)):"
+            if isinstance(module, HookPoint):
                 self.hook_dict[name] = module
 
     def hook_points(self):
         return self.hook_dict.values()
 
     def remove_all_hook_fns(
-        self, direction="both", including_permanent=False, level=None
+        self,
+        direction: Literal["fwd", "bwd", "both"] = "both",
+        including_permanent: bool = False,
+        level: Union[int, None] = None,
     ):
         for hp in self.hook_points():
-            hp.remove_hooks(
-                direction, including_permanent=including_permanent, level=level
-            )
+            hp.remove_hooks(direction, including_permanent=including_permanent, level=level)
 
     def clear_contexts(self):
         for hp in self.hook_points():
@@ -191,10 +242,10 @@ class HookedRootModule(nn.Module):
 
     def reset_hooks(
         self,
-        clear_contexts=True,
-        direction="both",
-        including_permanent=False,
-        level=None,
+        clear_contexts: bool = True,
+        direction: Literal["fwd", "bwd", "both"] = "both",
+        including_permanent: bool = False,
+        level: Union[int, None] = None,
     ):
         if clear_contexts:
             self.clear_contexts()
@@ -203,15 +254,16 @@ class HookedRootModule(nn.Module):
 
     def check_and_add_hook(
         self,
-        hook_point,
-        hook_point_name,
-        hook,
-        dir="fwd",
-        is_permanent=False,
-        level=None,
-        prepend=False,
+        hook_point: HookPoint,
+        hook_point_name: str,
+        hook: HookFunction,
+        dir: Literal["fwd", "bwd"] = "fwd",
+        is_permanent: bool = False,
+        level: Union[int, None] = None,
+        prepend: bool = False,
     ) -> None:
         """Runs checks on the hook, and then adds it to the hook point"""
+
         self.check_hooks_to_add(
             hook_point,
             hook_point_name,
@@ -220,22 +272,36 @@ class HookedRootModule(nn.Module):
             is_permanent=is_permanent,
             prepend=prepend,
         )
-        hook_point.add_hook(
-            hook, dir=dir, is_permanent=is_permanent, level=level, prepend=prepend
-        )
+        hook_point.add_hook(hook, dir=dir, is_permanent=is_permanent, level=level, prepend=prepend)
 
     def check_hooks_to_add(
-        self, hook_point, hook_point_name, hook, dir="fwd", is_permanent=False
+        self,
+        hook_point: HookPoint,
+        hook_point_name: str,
+        hook: HookFunction,
+        dir: Literal["fwd", "bwd"] = "fwd",
+        is_permanent: bool = False,
+        prepend: bool = False,
     ) -> None:
         """Override this function to add checks on which hooks should be added"""
         pass
 
     def add_hook(
-        self, name, hook, dir="fwd", is_permanent=False, level=None, prepend=False
+        self,
+        name: Union[str, Callable[[str], bool]],
+        hook: HookFunction,
+        dir: Literal["fwd", "bwd"] = "fwd",
+        is_permanent: bool = False,
+        level: Union[int, None] = None,
+        prepend: bool = False,
     ) -> None:
-        if type(name) == str:
+        if isinstance(name, str):
+            hook_point = self.mod_dict[name]
+            assert isinstance(
+                hook_point, HookPoint
+            )  # TODO does adding assert meaningfully slow down performance? I've added them for type checking purposes.
             self.check_and_add_hook(
-                self.mod_dict[name],
+                hook_point,
                 name,
                 hook,
                 dir=dir,
@@ -257,9 +323,58 @@ class HookedRootModule(nn.Module):
                         prepend=prepend,
                     )
 
-    def add_perma_hook(self, name, hook, dir="fwd") -> None:
+    def add_perma_hook(
+        self,
+        name: Union[str, Callable[[str], bool]],
+        hook: HookFunction,
+        dir: Literal["fwd", "bwd"] = "fwd",
+    ) -> None:
         self.add_hook(name, hook, dir=dir, is_permanent=True)
 
+    def _enable_hook_with_name(self, name: str, hook: Callable, dir: Literal["fwd", "bwd"]):
+        """This function takes a key for the mod_dict and enables the related hook for that module
+
+        Args:
+            name (str): The module name
+            hook (Callable): The hook to add
+            dir (Literal[&quot;fwd&quot;, &quot;bwd&quot;]): The direction for the hook
+        """
+        self.mod_dict[name].add_hook(hook, dir=dir, level=self.context_level)
+
+    def _enable_hooks_for_points(
+        self,
+        hook_points: Iterable[Tuple[str, HookPoint]],
+        enabled: Callable,
+        hook: Callable,
+        dir: Literal["fwd", "bwd"],
+    ):
+        """Enables hooks for a list of points
+
+        Args:
+            hook_points (Dict[str, HookPoint]): The hook points
+            enabled (Callable): _description_
+            hook (Callable): _description_
+            dir (Literal[&quot;fwd&quot;, &quot;bwd&quot;]): _description_
+        """
+        for hook_name, hook_point in hook_points:
+            if enabled(hook_name):
+                hook_point.add_hook(hook, dir=dir, level=self.context_level)
+
+    def _enable_hook(self, name: Union[str, Callable], hook: Callable, dir: Literal["fwd", "bwd"]):
+        """Enables an individual hook on a hook point
+
+        Args:
+            name (str): The name of the hook
+            hook (Callable): The actual hook
+            dir (Literal[&quot;fwd&quot;, &quot;bwd&quot;], optional): The direction of the hook. Defaults to "fwd".
+        """
+        if isinstance(name, str):
+            self._enable_hook_with_name(name=name, hook=hook, dir=dir)
+        else:
+            self._enable_hooks_for_points(
+                hook_points=self.hook_dict.items(), enabled=name, hook=hook, dir=dir
+            )
+
     @contextmanager
     def hooks(
         self,
@@ -279,35 +394,19 @@ class HookedRootModule(nn.Module):
             clear_contexts (bool): If True, clears hook contexts whenever hooks are reset.
 
         Example:
-        --------
+
         .. code-block:: python
 
-            >>> with model.hooks(fwd_hooks=my_hooks):
-            >>>     hooked_loss = model(text, return_type="loss")
+            with model.hooks(fwd_hooks=my_hooks):
+                hooked_loss = model(text, return_type="loss")
         """
         try:
             self.context_level += 1
 
             for name, hook in fwd_hooks:
-                if type(name) == str:
-                    self.mod_dict[name].add_hook(
-                        hook, dir="fwd", level=self.context_level
-                    )
-                else:
-                    # Otherwise, name is a Boolean function on names
-                    for hook_name, hp in self.hook_dict.items():
-                        if name(hook_name):
-                            hp.add_hook(hook, dir="fwd", level=self.context_level)
+                self._enable_hook(name=name, hook=hook, dir="fwd")
             for name, hook in bwd_hooks:
-                if type(name) == str:
-                    self.mod_dict[name].add_hook(
-                        hook, dir="bwd", level=self.context_level
-                    )
-                else:
-                    # Otherwise, name is a Boolean function on names
-                    for hook_name, hp in self.hook_dict:
-                        if name(hook_name):
-                            hp.add_hook(hook, dir="bwd", level=self.context_level)
+                self._enable_hook(name=name, hook=hook, dir="bwd")
             yield self
         finally:
             if reset_hooks_end:
@@ -318,12 +417,12 @@ class HookedRootModule(nn.Module):
 
     def run_with_hooks(
         self,
-        *model_args,
+        *model_args: Any,  # TODO: unsure about whether or not this Any typing is correct or not; may need to be replaced with something more specific?
         fwd_hooks: List[Tuple[Union[str, Callable], Callable]] = [],
         bwd_hooks: List[Tuple[Union[str, Callable], Callable]] = [],
-        reset_hooks_end=True,
-        clear_contexts=False,
-        **model_kwargs,
+        reset_hooks_end: bool = True,
+        clear_contexts: bool = False,
+        **model_kwargs: Any,
     ):
         """
         Runs the model with specified forward and backward hooks.
@@ -351,16 +450,14 @@ class HookedRootModule(nn.Module):
                 "WARNING: Hooks will be reset at the end of run_with_hooks. This removes the backward hooks before a backward pass can occur."
             )
 
-        with self.hooks(
-            fwd_hooks, bwd_hooks, reset_hooks_end, clear_contexts
-        ) as hooked_model:
+        with self.hooks(fwd_hooks, bwd_hooks, reset_hooks_end, clear_contexts) as hooked_model:
             return hooked_model.forward(*model_args, **model_kwargs)
 
     def add_caching_hooks(
         self,
         names_filter: NamesFilter = None,
         incl_bwd: bool = False,
-        device=None,
+        device: DeviceType = None,  # TODO: unsure about whether or not this device typing is correct or not?
         remove_batch_dim: bool = False,
         cache: Optional[dict] = None,
     ) -> dict:
@@ -381,44 +478,45 @@ class HookedRootModule(nn.Module):
 
         if names_filter is None:
             names_filter = lambda name: True
-        elif type(names_filter) == str:
+        elif isinstance(names_filter, str):
             filter_str = names_filter
             names_filter = lambda name: name == filter_str
-        elif type(names_filter) == list:
+        elif isinstance(names_filter, list):
             filter_list = names_filter
             names_filter = lambda name: name in filter_list
 
+        assert callable(names_filter), "names_filter must be a callable"
+
         self.is_caching = True
 
-        def save_hook(tensor, hook):
+        def save_hook(tensor: torch.Tensor, hook: HookPoint, is_backward: bool):
+            assert hook.name is not None
+            hook_name = hook.name
+            if is_backward:
+                hook_name += "_grad"
             if remove_batch_dim:
-                cache[hook.name] = tensor.detach().to(device)[0]
+                cache[hook_name] = tensor.detach().to(device)[0]
             else:
-                cache[hook.name] = tensor.detach().to(device)
-
-        def save_hook_back(tensor, hook):
-            if remove_batch_dim:
-                cache[hook.name + "_grad"] = tensor.detach().to(device)[0]
-            else:
-                cache[hook.name + "_grad"] = tensor.detach().to(device)
+                cache[hook_name] = tensor.detach().to(device)
 
         for name, hp in self.hook_dict.items():
             if names_filter(name):
-                hp.add_hook(save_hook, "fwd")
+                hp.add_hook(partial(save_hook, is_backward=False), "fwd")
                 if incl_bwd:
-                    hp.add_hook(save_hook_back, "bwd")
+                    hp.add_hook(partial(save_hook, is_backward=True), "bwd")
         return cache
 
     def run_with_cache(
         self,
-        *model_args,
+        *model_args: Any,
         names_filter: NamesFilter = None,
-        device=None,
-        remove_batch_dim=False,
-        incl_bwd=False,
-        reset_hooks_end=True,
-        clear_contexts=False,
-        **model_kwargs,
+        device: DeviceType = None,
+        remove_batch_dim: bool = False,
+        incl_bwd: bool = False,
+        reset_hooks_end: bool = True,
+        clear_contexts: bool = False,
+        pos_slice: Optional[Union[Slice, SliceInput]] = None,
+        **model_kwargs: Any,
     ):
         """
         Runs the model and returns the model output and a Cache object.
@@ -440,14 +538,23 @@ class HookedRootModule(nn.Module):
                 end of the run. Defaults to True.
             clear_contexts (bool, optional): If True, clears hook contexts whenever hooks are reset.
                 Defaults to False.
+            pos_slice:
+                The slice to apply to the cache output. Defaults to None, do nothing.
             **model_kwargs: Keyword arguments for the model.
 
         Returns:
             tuple: A tuple containing the model output and a Cache object.
 
         """
+
+        pos_slice = Slice.unwrap(pos_slice)
+
         cache_dict, fwd, bwd = self.get_caching_hooks(
-            names_filter, incl_bwd, device, remove_batch_dim=remove_batch_dim
+            names_filter,
+            incl_bwd,
+            device,
+            remove_batch_dim=remove_batch_dim,
+            pos_slice=pos_slice,
         )
 
         with self.hooks(
@@ -466,9 +573,10 @@ class HookedRootModule(nn.Module):
         self,
         names_filter: NamesFilter = None,
         incl_bwd: bool = False,
-        device=None,
+        device: DeviceType = None,
         remove_batch_dim: bool = False,
         cache: Optional[dict] = None,
+        pos_slice: Union[Slice, SliceInput] = None,
     ) -> Tuple[dict, list, list]:
         """Creates hooks to cache activations. Note: It does not add the hooks to the model.
 
@@ -487,39 +595,72 @@ class HookedRootModule(nn.Module):
         if cache is None:
             cache = {}
 
+        pos_slice = Slice.unwrap(pos_slice)
+
         if names_filter is None:
             names_filter = lambda name: True
-        elif type(names_filter) == str:
+        elif isinstance(names_filter, str):
             filter_str = names_filter
             names_filter = lambda name: name == filter_str
-        elif type(names_filter) == list:
+        elif isinstance(names_filter, list):
             filter_list = names_filter
             names_filter = lambda name: name in filter_list
+        elif callable(names_filter):
+            names_filter = names_filter
+        else:
+            raise ValueError("names_filter must be a string, list of strings, or function")
+        assert callable(names_filter)  # Callable[[str], bool]
+
         self.is_caching = True
 
-        def save_hook(tensor, hook):
-            if remove_batch_dim:
-                cache[hook.name] = tensor.detach().to(device)[0]
-            else:
-                cache[hook.name] = tensor.detach().to(device)
+        def save_hook(tensor: torch.Tensor, hook: HookPoint, is_backward: bool = False):
+            # for attention heads the pos dimension is the third from last
+            if hook.name is None:
+                raise RuntimeError("Hook should have been provided a name")
 
-        def save_hook_back(tensor, hook):
+            hook_name = hook.name
+            if is_backward:
+                hook_name += "_grad"
+            resid_stream = tensor.detach().to(device)
             if remove_batch_dim:
-                cache[hook.name + "_grad"] = tensor.detach().to(device)[0]
+                resid_stream = resid_stream[0]
+
+            if (
+                hook.name.endswith("hook_q")
+                or hook.name.endswith("hook_k")
+                or hook.name.endswith("hook_v")
+                or hook.name.endswith("hook_z")
+                or hook.name.endswith("hook_result")
+            ):
+                pos_dim = -3
             else:
-                cache[hook.name + "_grad"] = tensor.detach().to(device)
+                # for all other components the pos dimension is the second from last
+                # including the attn scores where the dest token is the second from last
+                pos_dim = -2
+
+            if (
+                tensor.dim() >= -pos_dim
+            ):  # check if the residual stream has a pos dimension before trying to slice
+                resid_stream = pos_slice.apply(resid_stream, dim=pos_dim)
+            cache[hook_name] = resid_stream
 
         fwd_hooks = []
         bwd_hooks = []
-        for name, hp in self.hook_dict.items():
+        for name, _ in self.hook_dict.items():
             if names_filter(name):
-                fwd_hooks.append((name, save_hook))
+                fwd_hooks.append((name, partial(save_hook, is_backward=False)))
                 if incl_bwd:
-                    bwd_hooks.append((name, save_hook_back))
+                    bwd_hooks.append((name, partial(save_hook, is_backward=True)))
 
         return cache, fwd_hooks, bwd_hooks
 
-    def cache_all(self, cache, incl_bwd=False, device=None, remove_batch_dim=False):
+    def cache_all(
+        self,
+        cache: Optional[dict],
+        incl_bwd: bool = False,
+        device: DeviceType = None,
+        remove_batch_dim: bool = False,
+    ):
         logging.warning(
             "cache_all is deprecated and will eventually be removed, use add_caching_hooks or run_with_cache"
         )
@@ -533,11 +674,11 @@ class HookedRootModule(nn.Module):
 
     def cache_some(
         self,
-        cache,
+        cache: Optional[dict],
         names: Callable[[str], bool],
-        incl_bwd=False,
-        device=None,
-        remove_batch_dim=False,
+        incl_bwd: bool = False,
+        device: DeviceType = None,
+        remove_batch_dim: bool = False,
     ):
         """Cache a list of hook provided by names, Boolean function on names"""
         logging.warning(
