@@ -1,14 +1,22 @@
+"""Utils.
+
+This module contains varied utility functions used throughout the library.
+"""
+
 from __future__ import annotations
 
 import inspect
 import json
+import os
 import re
 import shutil
-from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union, cast
+from copy import deepcopy
+from typing import Any, Callable, Dict, List, Optional, Tuple, Union, cast
 
 import einops
 import numpy as np
 import torch
+import torch.nn as nn
 import torch.nn.functional as F
 import transformers
 from datasets.arrow_dataset import Dataset
@@ -18,20 +26,15 @@ from jaxtyping import Float, Int
 from rich import print as rprint
 from transformers import AutoTokenizer
 
-from transformer_lens import FactoredMatrix
+from transformer_lens.FactoredMatrix import FactoredMatrix
 
 CACHE_DIR = transformers.TRANSFORMERS_CACHE
+USE_DEFAULT_VALUE = None
 
 
-def select_compatible_kwargs(
-    kwargs_dict: Dict[str, Any], callable: Callable
-) -> Dict[str, Any]:
+def select_compatible_kwargs(kwargs_dict: Dict[str, Any], callable: Callable) -> Dict[str, Any]:
     """Return a dict with the elements kwargs_dict that are parameters of callable"""
-    return {
-        k: v
-        for k, v in kwargs_dict.items()
-        if k in inspect.getfullargspec(callable).args
-    }
+    return {k: v for k, v in kwargs_dict.items() if k in inspect.getfullargspec(callable).args}
 
 
 def download_file_from_hf(
@@ -81,9 +84,7 @@ def clear_huggingface_cache():
 
 
 def print_gpu_mem(step_name=""):
-    print(
-        f"{step_name} ~ {np.round(torch.cuda.memory_allocated()/2e30, 2)} GiB allocated on GPU."
-    )
+    print(f"{step_name} ~ {np.round(torch.cuda.memory_allocated()/2e30, 2)} GiB allocated on GPU.")
 
 
 def get_corner(tensor, n=3):
@@ -127,9 +128,7 @@ def lm_cross_entropy_loss(
     # Use torch.gather to find the log probs of the correct tokens
     # Offsets needed because we're predicting the NEXT token (this means the final logit is meaningless)
     # None and [..., 0] needed because the tensor used in gather must have the same rank.
-    predicted_log_probs = log_probs[..., :-1, :].gather(
-        dim=-1, index=tokens[..., 1:, None]
-    )[..., 0]
+    predicted_log_probs = log_probs[..., :-1, :].gather(dim=-1, index=tokens[..., 1:, None])[..., 0]
     if per_token:
         return -predicted_log_probs
     else:
@@ -160,28 +159,17 @@ def gelu_new(
     return (
         0.5
         * input
-        * (
-            1.0
-            + torch.tanh(
-                np.sqrt(2.0 / np.pi) * (input + 0.044715 * torch.pow(input, 3.0))
-            )
-        )
+        * (1.0 + torch.tanh(np.sqrt(2.0 / np.pi) * (input + 0.044715 * torch.pow(input, 3.0))))
     )
 
 
 def gelu_fast(
     input: Float[torch.Tensor, "batch pos d_mlp"]
 ) -> Float[torch.Tensor, "batch pos d_mlp"]:
-    return (
-        0.5
-        * input
-        * (1.0 + torch.tanh(input * 0.7978845608 * (1.0 + 0.044715 * input * input)))
-    )
+    return 0.5 * input * (1.0 + torch.tanh(input * 0.7978845608 * (1.0 + 0.044715 * input * input)))
 
 
-def solu(
-    input: Float[torch.Tensor, "batch pos d_mlp"]
-) -> Float[torch.Tensor, "batch pos d_mlp"]:
+def solu(input: Float[torch.Tensor, "batch pos d_mlp"]) -> Float[torch.Tensor, "batch pos d_mlp"]:
     """
     SoLU activation function as described by
     https://transformer-circuits.pub/2022/solu/index.html.
@@ -191,6 +179,81 @@ def solu(
     return input * F.softmax(input, dim=-1)
 
 
+def calc_fan_in_and_fan_out(tensor):
+    """
+    Calculate the fan in and fan out of a tensor. We define it ourselves because Torch uses a
+    different convention for weights (e.g. for an MLP they use d_out x d_in, and we use d_in x
+    d_out, for attention they do (n_head d_head) x d_model, we do n_head x d_model x d_head).
+    """
+    shape = tensor.shape
+
+    if len(shape) == 0:
+        raise ValueError("Fan in and fan out can not be computed for scalars.")
+    elif len(shape) == 1:
+        fan_in = 1
+        fan_out = shape[0]
+    elif len(shape) == 2:  # Linear transform
+        fan_in = shape[0]
+        fan_out = shape[1]
+    elif len(shape) == 3:  # Attention head weight, has shape n_head x d_model x d_head
+        fan_in = shape[1]
+        fan_out = shape[0] * shape[2]
+    else:
+        raise ValueError(f"Fan in and fan out can not be computed for shape {shape} tensors.")
+
+    return fan_in, fan_out
+
+
+def init_xavier_uniform_(param, gain=1.0):
+    """
+    Initializes the input tensor using the Xavier initialization method.
+    """
+    fan_in, fan_out = calc_fan_in_and_fan_out(param)
+    max = gain * np.sqrt(6.0 / (fan_in + fan_out))
+    return nn.init.uniform_(param, -max, max)
+
+
+def init_xavier_normal_(param, gain=1.0):
+    """
+    Initializes the input tensor using the Xavier initialization method.
+    """
+    fan_in, fan_out = calc_fan_in_and_fan_out(param)
+    std = gain * np.sqrt(2.0 / (fan_in + fan_out))
+    return nn.init.normal_(param, mean=0.0, std=std)
+
+
+def init_kaiming_uniform_(param, a=0, nonlinearity="relu", gain=1.0, mode="fan_in"):
+    """
+    Initializes the input tensor using the Kaiming initialization method.
+
+    Starting from a std 1 uniform distribution, we scale the weights by c / sqrt(fan_in), where c =
+    sqrt(2) if the params were immediately preceded by a relu and 1 for everything else.
+
+    As with torch, `a` is a hyperparameter for `nonlinearity`, if it takes one.
+    """
+    fan_in, fan_out = calc_fan_in_and_fan_out(param)
+    fan = fan_in if mode == "fan_in" else fan_out
+    gain *= nn.init.calculate_gain(nonlinearity, a)
+    max = gain * np.sqrt(3.0 / fan)
+    return nn.init.uniform_(param, -max, max)
+
+
+def init_kaiming_normal_(param, a=0, nonlinearity="relu", gain=1.0, mode="fan_in"):
+    """
+    Initializes the input tensor using the Kaiming initialization method.
+
+    Starting from a std 1 normal distribution, we scale the weights by c / sqrt(fan_in), where c =
+    sqrt(2) if the params were immediately preceded by a relu and 1 for everything else.
+
+    As with torch, `a` is a hyperparameter for `nonlinearity`, if it takes one.
+    """
+    fan_in, fan_out = calc_fan_in_and_fan_out(param)
+    fan = fan_in if mode == "fan_in" else fan_out
+    gain *= nn.init.calculate_gain(nonlinearity, a)
+    std = gain * np.sqrt(1.0 / fan)
+    return nn.init.normal_(param, mean=0.0, std=std)
+
+
 def keep_single_column(dataset: Dataset, col_name: str):
     """
     Acts on a HuggingFace dataset to delete all columns apart from a single column name - useful when we want to tokenize and mix together different strings
@@ -244,14 +307,9 @@ def tokenize_and_concatenate(
         # Divide into 20 chunks of ~ equal length
         num_chunks = 20
         chunk_length = (len(full_text) - 1) // num_chunks + 1
-        chunks = [
-            full_text[i * chunk_length : (i + 1) * chunk_length]
-            for i in range(num_chunks)
-        ]
+        chunks = [full_text[i * chunk_length : (i + 1) * chunk_length] for i in range(num_chunks)]
         # Tokenize the chunks in parallel. Uses NumPy because HuggingFace map doesn't want tensors returned
-        tokens = tokenizer(chunks, return_tensors="np", padding=True)[
-            "input_ids"
-        ].flatten()
+        tokens = tokenizer(chunks, return_tensors="np", padding=True)["input_ids"].flatten()
         # Drop padding tokens
         tokens = tokens[tokens != tokenizer.pad_token_id]
         num_tokens = len(tokens)
@@ -276,16 +334,6 @@ def tokenize_and_concatenate(
     return tokenized_dataset
 
 
-""" 
-Test ^
-
-data = Dataset.from_dict({"text":[str(i) for i in range(1000)]})
-tokenizer = AutoTokenizer.from_pretrained("NeelNanda/gpt-neox-tokenizer-digits")
-print(data)
-tokenize_and_concatenate(data, tokenizer, streaming=False, column_name="text")
-"""
-
-
 def sample_logits(
     final_logits: Float[torch.Tensor, "batch d_vocab"],
     top_k: Optional[int] = None,
@@ -316,9 +364,7 @@ def sample_logits(
 
         final_logits = final_logits / temperature
         if freq_penalty > 0:
-            assert (
-                tokens is not None
-            ), "Must provide input_tokens if applying a frequency penalty"
+            assert tokens is not None, "Must provide input_tokens if applying a frequency penalty"
             for batch_index in range(final_logits.shape[0]):
                 # torch.bincount returns a tensor of length d_vocab, with the number of occurences of each token in the tokens.
                 final_logits[batch_index] = final_logits[
@@ -337,9 +383,7 @@ def sample_logits(
             cumulative_probs = sorted_logits.softmax(dim=-1).cumsum(dim=-1)
             # We round up - we want prob >= top_p not <top_p
             sorted_indices_to_remove = cumulative_probs > top_p
-            sorted_indices_to_remove[..., 1:] = sorted_indices_to_remove[
-                ..., :-1
-            ].clone()
+            sorted_indices_to_remove[..., 1:] = sorted_indices_to_remove[..., :-1].clone()
             sorted_indices_to_remove[..., 0] = 0
             indices_to_remove = sorted_indices_to_remove.scatter(
                 -1, sorted_indices, sorted_indices_to_remove
@@ -350,9 +394,8 @@ def sample_logits(
         return torch.distributions.categorical.Categorical(logits=final_logits).sample()
 
 
-# %%
 # Type alias
-SliceInput: Type = Optional[
+SliceInput = Optional[
     Union[
         int,
         Tuple[int,],
@@ -363,7 +406,8 @@ SliceInput: Type = Optional[
         np.ndarray,
     ]
 ]
-"""
+"""An object that represents a slice input. It can be a tuple of integers or a slice object.
+
 An optional type alias for a slice input used in the `ActivationCache` module.
 
 A `SliceInput` can be one of the following types:
@@ -374,14 +418,12 @@ A `SliceInput` can be one of the following types:
     - `torch.Tensor`: a tensor containing a boolean mask or a list of indices to be selected from the input tensor.
 
 `SliceInput` is used in the `apply_ln_to_stack` method in the `ActivationCache` module.
-
-:class:`SliceInput`
-    An object that represents a slice input. It can be a tuple of integers or a slice object.
 """
 
 
 class Slice:
-    """
+    """An object that represents a slice input. It can be a tuple of integers or a slice object.
+
     We use a custom slice syntax because Python/Torch's don't let us reduce the number of dimensions:
 
     Note that slicing with input_slice=None means do nothing, NOT add an extra dimension (use unsqueeze for that)
@@ -398,11 +440,10 @@ class Slice:
     elif input_slice = (1, 5, 2), tensor -> tensor[1:5:2] (ie indexing with [1, 3])
     elif input_slice = [1, 4, 5], tensor -> tensor[[1, 4, 5]] (ie changing the first axis to have length 3, and taking the indices 1, 4, 5 out).
     elif input_slice is a Tensor, same as list - Tensor is assumed to be a 1D list of indices.
-
-    :class: `Slice`
-        An object that represents a slice input. It can be a tuple of integers or a slice object.
     """
 
+    slice: Union[int, slice, np.ndarray]
+
     def __init__(
         self,
         input_slice: SliceInput = None,
@@ -413,20 +454,16 @@ class Slice:
         Args:
             input_slice (SliceInput): The slice to apply. Can be an int, a tuple, a list, a torch.Tensor, or None. If None, do nothing.
 
-        Returns:
-            Slice: A Slice object that can be applied to a tensor.
-
         Raises:
             ValueError: If the input_slice is not one of the above types.
         """
-        if type(input_slice) == tuple:
-            input_slice: slice = slice(*input_slice)
-            self.slice = input_slice
+        if isinstance(input_slice, tuple):
+            self.slice = slice(*input_slice)
             self.mode = "slice"
-        elif type(input_slice) == int:
+        elif isinstance(input_slice, int):
             self.slice = input_slice
             self.mode = "int"
-        elif type(input_slice) == slice:
+        elif isinstance(input_slice, slice):
             self.slice = input_slice
             self.mode = "slice"
         elif type(input_slice) in [list, torch.Tensor, np.ndarray]:
@@ -455,7 +492,7 @@ class Slice:
         """
         ndim = tensor.ndim
         slices = [slice(None)] * ndim
-        slices[dim] = self.slice
+        slices[dim] = self.slice  # type: ignore
         return tensor[tuple(slices)]
 
     def indices(
@@ -485,13 +522,32 @@ class Slice:
     ) -> str:
         return f"Slice: {self.slice} Mode: {self.mode} "
 
+    @classmethod
+    def unwrap(
+        cls,
+        slice_input: Union["Slice", SliceInput],
+    ) -> "Slice":
+        """
+        Takes a Slice-like input and converts it into a Slice, if it is not already.
 
-# %%
+        Args:
+            slice_input (Union[Slice, SliceInput]): The input to turn into a Slice.
+
+        Returns:
+            Slice: A Slice object.
+        """
+        if not isinstance(slice_input, Slice):
+            if isinstance(
+                slice_input, int
+            ):  # slicing with an int collapses the dimension so this stops the pos dimension from collapsing
+                slice_input = [slice_input]
+            slice_input = Slice(slice_input)
+        return slice_input
 
 
 def get_act_name(
     name: str,
-    layer: Optional[int] = None,
+    layer: Optional[Union[int, str]] = None,
     layer_type: Optional[str] = None,
 ):
     """
@@ -527,16 +583,12 @@ def get_act_name(
     get_act_name('scale4ln1')=='blocks.4.ln1.hook_scale'
     get_act_name('pre5')=='blocks.5.mlp.hook_pre'
     """
-    if (
-        ("." in name or name.startswith("hook_"))
-        and layer is None
-        and layer_type is None
-    ):
+    if ("." in name or name.startswith("hook_")) and layer is None and layer_type is None:
         # If this was called on a full name, just return it
         return name
     match = re.match(r"([a-z]+)(\d+)([a-z]?.*)", name)
     if match is not None:
-        name, layer, layer_type = match.groups(0)
+        name, layer, layer_type = match.groups(0)  # type: ignore
 
     layer_type_alias = {
         "a": "attn",
@@ -578,7 +630,7 @@ def get_act_name(
         "attn_scores",
     ]:
         layer_type = "attn"
-    elif name in ["pre", "post", "mid"]:
+    elif name in ["pre", "post", "mid", "pre_linear"]:
         layer_type = "mlp"
     elif layer_type in layer_type_alias:
         layer_type = layer_type_alias[layer_type]
@@ -592,9 +644,7 @@ def get_act_name(
     return full_act_name
 
 
-def remove_batch_dim(
-    tensor: Float[torch.Tensor, "1 ..."]
-) -> Float[torch.Tensor, "..."]:
+def remove_batch_dim(tensor: Float[torch.Tensor, "1 ..."]) -> Float[torch.Tensor, "..."]:
     """
     Removes the first dimension of a tensor if it is size 1, otherwise returns the tensor unchanged
     """
@@ -607,23 +657,78 @@ def remove_batch_dim(
 def test_prompt(
     prompt: str,
     answer: str,
-    model,
+    model,  # Can't give type hint due to circular imports
     prepend_space_to_answer: bool = True,
     print_details: bool = True,
-    prepend_bos: Optional[bool] = None,
+    prepend_bos: Optional[bool] = USE_DEFAULT_VALUE,
     top_k: int = 10,
-):
-    """
-    Function to test whether a model can give the correct answer to a prompt. Intended for exploratory analysis, so it prints things out rather than returning things.
+) -> None:
+    """Test if the Model Can Give the Correct Answer to a Prompt.
 
-    Works for multi-token answers and multi-token prompts.
+    Intended for exploratory analysis. Prints out the performance on the answer (rank, logit, prob),
+    as well as the top k tokens. Works for multi-token prompts and multi-token answers.
 
-    Will always print the ranks of the answer tokens, and if print_details will print the logit and prob for the answer tokens and the top k tokens returned for each answer position.
+    Warning:
+
+    This will print the results (it does not return them).
+
+    Examples:
+
+    >>> from transformer_lens import HookedTransformer, utils
+    >>> model = HookedTransformer.from_pretrained("tiny-stories-1M")
+    Loaded pretrained model tiny-stories-1M into HookedTransformer
+
+    >>> prompt = "Why did the elephant cross the"
+    >>> answer = "road"
+    >>> utils.test_prompt(prompt, answer, model)
+    Tokenized prompt: ['<|endoftext|>', 'Why', ' did', ' the', ' elephant', ' cross', ' the']
+    Tokenized answer: [' road']
+    Performance on answer token:
+    Rank: 2        Logit: 14.24 Prob:  3.51% Token: | road|
+    Top 0th token. Logit: 14.51 Prob:  4.59% Token: | ground|
+    Top 1th token. Logit: 14.41 Prob:  4.18% Token: | tree|
+    Top 2th token. Logit: 14.24 Prob:  3.51% Token: | road|
+    Top 3th token. Logit: 14.22 Prob:  3.45% Token: | car|
+    Top 4th token. Logit: 13.92 Prob:  2.55% Token: | river|
+    Top 5th token. Logit: 13.79 Prob:  2.25% Token: | street|
+    Top 6th token. Logit: 13.77 Prob:  2.21% Token: | k|
+    Top 7th token. Logit: 13.75 Prob:  2.16% Token: | hill|
+    Top 8th token. Logit: 13.64 Prob:  1.92% Token: | swing|
+    Top 9th token. Logit: 13.46 Prob:  1.61% Token: | park|
+    Ranks of the answer tokens: [(' road', 2)]
+
+    Args:
+        prompt:
+            The prompt string, e.g. "Why did the elephant cross the".
+        answer:
+            The answer, e.g. "road". Note that if you set prepend_space_to_answer to False, you need
+            to think about if you have a space before the answer here (as e.g. in this example the
+            answer may really be " road" if the prompt ends without a trailing space).
+        model:
+            The model.
+        prepend_space_to_answer:
+            Whether or not to prepend a space to the answer. Note this will only ever prepend a
+            space if the answer doesn't already start with one.
+        print_details:
+            Print the prompt (as a string but broken up by token), answer and top k tokens (all
+            with logit, rank and probability).
+        prepend_bos:
+            Overrides self.cfg.default_prepend_bos if set. Whether to prepend
+            the BOS token to the input (applicable when input is a string). Models generally learn
+            to use the BOS token as a resting place for attention heads (i.e. a way for them to be
+            "turned off"). This therefore often improves performance slightly.
+        top_k:
+            Top k tokens to print details of (when print_details is set to True).
+
+    Returns:
+        None (just prints the results directly).
     """
     if prepend_space_to_answer and not answer.startswith(" "):
         answer = " " + answer
     # GPT-2 often treats the first token weirdly, so lets give it a resting position
-    tokens = model.to_tokens(prompt + answer, prepend_bos=prepend_bos)
+    prompt_tokens = model.to_tokens(prompt, prepend_bos=prepend_bos)
+    answer_tokens = model.to_tokens(answer, prepend_bos=False)
+    tokens = torch.cat((prompt_tokens, answer_tokens), dim=1)
     prompt_str_tokens = model.to_str_tokens(prompt, prepend_bos=prepend_bos)
     answer_str_tokens = model.to_str_tokens(answer, prepend_bos=False)
     prompt_length = len(prompt_str_tokens)
@@ -658,7 +763,6 @@ def test_prompt(
     rprint(f"[b]Ranks of the answer tokens:[/b] {answer_ranks}")
 
 
-# %%
 def transpose(tensor: Float[torch.Tensor, "... a b"]) -> Float[torch.Tensor, "... b a"]:
     """
     Utility to swap the last two dimensions of a tensor, regardless of the number of leading dimensions
@@ -686,15 +790,14 @@ def composition_scores(
         left.rdim == right.ldim
     ), f"Composition scores require left.rdim==right.ldim, shapes were left: {left.shape}, right:{right.shape}"
 
-    right = right.collapse_r()
-    left = left.collapse_l()
-    r_norms = right.norm(dim=[-2, -1])
-    l_norms = left.norm(dim=[-2, -1])
-    comp_norms = (left @ right).norm(dim=[-2, -1])
+    new_right = right.collapse_r()
+    new_left = left.collapse_l()
+    r_norms = new_right.norm(dim=[-2, -1])
+    l_norms = new_left.norm(dim=[-2, -1])
+    comp_norms = (new_left @ new_right).norm(dim=[-2, -1])
     return comp_norms / r_norms / l_norms
 
 
-# %%
 def get_dataset(dataset_name: str, **kwargs) -> Dataset:
     """
     Returns a small HuggingFace dataset, for easy testing and exploration. Accesses several convenience datasets with 10,000 elements (dealing with the enormous 100GB - 2TB datasets is a lot of effort!). Note that it returns a dataset (ie a dictionary containing all the data), *not* a DataLoader (iterator over the data + some fancy features). But you can easily convert it to a DataLoader.
@@ -741,9 +844,7 @@ def is_lower_triangular(x: torch.Tensor) -> bool:
     return x.equal(x.tril())
 
 
-def check_structure(
-    t1: torch.Tensor, t2: torch.Tensor, *, verbose: bool = False
-) -> None:
+def check_structure(t1: torch.Tensor, t2: torch.Tensor, *, verbose: bool = False) -> None:
     """Validate that the two square tensors have the same structure, i.e.,
     that the directionality of comparisons points in the same directions both
     row-wise and column-wise.
@@ -799,13 +900,313 @@ def get_device():
     return torch.device("cpu")
 
 
-def override_or_use_default_flag(
-    default_flag: bool,
-    override: Optional[bool] = None,
-) -> bool:
+def override_or_use_default_value(
+    default_flag: Any,
+    override: Optional[Any] = None,
+) -> Any:
     """
     Determines which flag to return based on whether an overriding flag is provided.
     If a not-None overriding flag is provided, it is returned.
     Otherwise, the global flag is returned.
     """
     return override if override is not None else default_flag
+
+
+def get_offset_position_ids(
+    past_kv_pos_offset: int,
+    attention_mask: Int[torch.Tensor, "batch offset_pos"],
+) -> Int[torch.Tensor, "batch pos"]:
+    """
+    Returns the indices of non-padded tokens, offset by the position of the first attended token.
+    """
+    # shift the position ids so that the id at the the first attended token position becomes zero.
+    # The position ids of the prepending pad tokens are shifted to -1.
+    shifted_position_ids = attention_mask.cumsum(dim=1) - 1  # [batch, tokens_length]
+
+    # Set the position ids of all prepending pad tokens to an arbitrary number (zero here)
+    # just to avoid indexing errors.
+    position_ids = shifted_position_ids.masked_fill(shifted_position_ids < 0, 0)
+    return position_ids[:, past_kv_pos_offset:]  # [pos, batch]
+
+
+def get_cumsum_along_dim(tensor, dim, reverse=False):
+    """
+    Returns the cumulative sum of a tensor along a given dimension.
+    """
+    if reverse:
+        tensor = tensor.flip(dims=(dim,))
+    cumsum = tensor.cumsum(dim=dim)
+    if reverse:
+        cumsum = cumsum.flip(dims=(dim,))
+    return cumsum
+
+
+def get_attention_mask(tokenizer, tokens: torch.Tensor, prepend_bos: bool) -> torch.Tensor:
+    """
+    Computes the attention mask for the tokenized input.
+    NOTE: Only the leftmost leading pads (when `padding_side == left`)
+    or rightmost trailing pads (when `padding_side == right`) are
+    considered as real pad tokens that should not be attended.
+
+    Args:
+        tokenizer: The tokenizer used for tokenization.
+        tokens (torch.Tensor): The tokenized input.
+        prepend_bos (bool): If True, a BOS token is prepended to the input.
+
+    Returns:
+        torch.Tensor: The attention mask for the input.
+    """
+
+    # Initialize the attention mask with ones (indicating all tokens should be attended to)
+    attention_mask = torch.ones_like(tokens)
+    is_not_pad_token = tokens.ne(tokenizer.pad_token_id)
+
+    if tokenizer.padding_side == "right":
+        # Zero-out the rightmost trailing pad tokens
+        is_trailing_pad = get_cumsum_along_dim(is_not_pad_token, -1, reverse=True) == 0
+        attention_mask[is_trailing_pad] = 0
+    else:
+        # Zero-out the leftmost leading pad tokens
+        is_leading_pad = get_cumsum_along_dim(is_not_pad_token, -1, reverse=False) == 0
+        attention_mask[is_leading_pad] = 0
+
+        # If the bos token is the same as the pad token,
+        # the last token of the leftmost leading pad tokens is the bos token.
+        # We need to set the attention mask for the bos token to 1.
+        if prepend_bos and tokenizer.bos_token_id == tokenizer.pad_token_id:
+            pad_bos_positions = is_leading_pad.sum(-1) - 1
+            attention_mask[torch.arange(attention_mask.shape[0]), pad_bos_positions] = 1
+
+    return attention_mask
+
+
+def repeat_along_head_dimension(
+    tensor: Float[torch.Tensor, "batch pos d_model"],
+    n_heads: int,
+    clone_tensor=True,
+    # `einops.repeat` uses a view in torch, so we generally clone the tensor to avoid using shared storage for each head entry
+):
+    repeated_tensor = einops.repeat(
+        tensor,
+        "batch pos d_model -> batch pos n_heads d_model",
+        n_heads=n_heads,
+    )
+    if clone_tensor:
+        return repeated_tensor.clone()
+    else:
+        return repeated_tensor
+
+
+def get_nested_attr(obj, attr_str):
+    """
+    Retrieves a nested attribute from an object based on a dot-separated string.
+
+    For example, if `attr_str` is "a.b.c", this function will return `obj.a.b.c`.
+
+    Args:
+        obj (Any): The object from which to retrieve the attribute.
+        attr_str (str): A dot-separated string representing the attribute hierarchy.
+
+    Returns:
+        Any: The value of the nested attribute.
+    """
+    attrs = attr_str.split(".")
+    for attr in attrs:
+        obj = getattr(obj, attr)
+    return obj
+
+
+def set_nested_attr(obj, attr_str, value):
+    """
+    Sets a nested attribute of an object based on a dot-separated string.
+
+    For example, if `attr_str` is "a.b.c", this function will set the value of `obj.a.b.c` to `value`.
+
+    Args:
+        obj (Any): The object on which to set the attribute.
+        attr_str (str): A dot-separated string representing the attribute hierarchy.
+        value (Any): The value to set for the nested attribute.
+    """
+    attrs = attr_str.split(".")
+
+    # Navigate to the deepest object containing the attribute to be set
+    for attr in attrs[:-1]:
+        obj = getattr(obj, attr)
+
+    # Set the nested attribute's value
+    setattr(obj, attrs[-1], value)
+
+
+class LocallyOverridenDefaults:
+    """
+    Context manager that allows temporary overriding of default values within a model.
+    Once the context is exited, the default values are restored.
+
+    WARNING: This context manager must be used for any function/method that directly accesses
+    default values which may be overridden by the user using the function/method's arguments,
+    e.g., `model.cfg.default_prepend_bos` and `model.tokenizer.padding_side` which can be
+    overriden by `prepend_bos` and `padding_side` arguments, respectively, in the `to_tokens`.
+    """
+
+    def __init__(self, model, **overrides):
+        """
+        Initializes the context manager.
+
+        Args:
+            model (HookedTransformer): The model whose default values will be overridden.
+            overrides (dict): Key-value pairs of properties to override and their new values.
+        """
+        self.model = model
+        self.overrides = overrides
+
+        # Dictionary defining valid defaults, valid values, and locations to find and store them
+        self.values_with_defaults = {
+            "prepend_bos": {
+                "default_location": "model.cfg.default_prepend_bos",
+                "valid_values": [USE_DEFAULT_VALUE, True, False],
+                "skip_overriding": False,
+                "default_value_to_restore": None,  # Will be set later
+            },
+            "padding_side": {
+                "default_location": "model.tokenizer.padding_side",
+                "valid_values": [USE_DEFAULT_VALUE, "left", "right"],
+                "skip_overriding": model.tokenizer is None,  # Do not override if tokenizer is None
+                "default_value_to_restore": None,  # Will be set later
+            },
+        }
+
+        # Ensure provided overrides are defined in the dictionary above
+        for override in overrides:
+            assert override in self.values_with_defaults, (
+                f"{override} is not a valid parameter to override. "
+                f"Valid parameters are {self.values_with_defaults.keys()}."
+            )
+
+    def __enter__(self):
+        """
+        Override default values upon entering the context.
+        """
+        for property, override in self.overrides.items():
+            info = self.values_with_defaults[property]
+            if info["skip_overriding"]:
+                continue  # Skip if overriding for this property is disabled
+
+            # Ensure the override is a valid value
+            valid_values = info["valid_values"]
+            assert (
+                override in valid_values  # type: ignore
+            ), f"{property} must be one of {valid_values}, but got {override}."
+
+            # Fetch current default and store it to restore later
+            default_location = info["default_location"]
+            default_value = get_nested_attr(self, default_location)
+            info["default_value_to_restore"] = deepcopy(default_value)
+
+            # Override the default value
+            locally_overriden_value = override_or_use_default_value(default_value, override)
+            set_nested_attr(self, default_location, locally_overriden_value)
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        """
+        Restore default values upon exiting the context.
+        """
+        for property in self.overrides:
+            info = self.values_with_defaults[property]
+            if info["skip_overriding"]:
+                continue
+
+            # Restore the default value from before the context was entered
+            default_location = info["default_location"]
+            default_value = info["default_value_to_restore"]
+            set_nested_attr(self, default_location, default_value)
+
+
+def get_tokenizer_with_bos(tokenizer):
+    """
+    Returns the tokenizer initialized with add_bos_token=True.
+    Such a tokenizer should be set as the default tokenizer because the tokenization of some
+    tokenizers like LlamaTokenizer are different when bos token is automatically/manually
+    prepended.
+
+    Args:
+        tokenizer (AutoTokenizer): The tokenizer to initialize with add_bos_token=True.
+
+    Returns:
+        AutoTokenizer: The tokenizer initialized with add_bos_token=True.
+    """
+    init_kwargs = deepcopy(tokenizer.init_kwargs)
+    pretrained_model_name_or_path = init_kwargs.pop("name_or_path")
+    add_bos_token = init_kwargs.pop("add_bos_token", None)
+    if add_bos_token is None:
+        add_bos_token = getattr(tokenizer, "add_bos_token", False)
+
+    if add_bos_token:
+        tokenizer_with_bos = tokenizer
+    else:
+        huggingface_token = os.environ.get("HF_TOKEN", None)
+        tokenizer_with_bos = AutoTokenizer.from_pretrained(
+            pretrained_model_name_or_path,
+            add_bos_token=True,
+            token=huggingface_token,
+            **init_kwargs,
+        )
+
+    return tokenizer_with_bos
+
+
+def get_input_with_manually_prepended_bos(tokenizer, input):
+    """
+    Manually prepends the bos token to the input.
+
+    Args:
+        tokenizer (AutoTokenizer): The tokenizer to use for prepending the bos token.
+        input (Union[str, List[str]]): The input to prepend the bos token to.
+
+    Returns:
+        Union[str, List[str]]: The input with the bos token manually prepended.
+    """
+    if isinstance(input, str):
+        input = tokenizer.bos_token + input
+    else:
+        input = [tokenizer.bos_token + string for string in input]
+    return input
+
+
+def get_tokens_with_bos_removed(tokenizer, tokens):
+    """
+    Removes the bos token from the beginning of each sequence in `tokens`.
+    The last dimension of `tokens` must be the sequence length.
+
+    Args:
+        tokenizer (AutoTokenizer): The tokenizer used to tokenize the input.
+        tokens (torch.Tensor): The tokenized input.
+
+    Returns:
+        torch.Tensor: The tokenized input with the bos token removed.
+    """
+    if tokenizer.padding_side == "right":
+        return tokens[..., 1:]
+
+    else:
+        bos_removed_shape = list(tokens.shape)
+        bos_removed_shape[-1] -= 1
+
+        if tokenizer.bos_token_id == tokenizer.pad_token_id:
+            is_not_pad_token = tokens.ne(tokenizer.pad_token_id)
+            is_leading_pad = get_cumsum_along_dim(is_not_pad_token, -1, reverse=False) == 0
+            real_bos_positions = is_leading_pad.sum(-1) - 1
+        else:
+            real_bos_positions = (tokens == tokenizer.bos_token_id).int().argmax(-1)
+
+        tokens = tokens.scatter(dim=1, index=real_bos_positions.unsqueeze(-1), value=-100)
+        return tokens[tokens != -100].view(*bos_removed_shape)
+
+
+try:
+    import pytest
+
+    # Note: Docstring won't be tested with PyTest (it's ignored), as it thinks this is a regular unit
+    # test (because its name is prefixed `test_`).
+    pytest.mark.skip(test_prompt)
+except ModuleNotFoundError:
+    pass  # disregard if pytest not in env
